/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * Copyright (C) 2015  HUBzero Foundation, LLC
 *
 * Author: Leif Delgass <ldelgass@purdue.edu>
 */
#include <vector>
#include <cstdio>
#include <cassert>

#include <osgEarth/Registry>
#include <osgEarth/ShaderGenerator>
#include <osgEarth/ObjectIndex>
#include <osgEarthFeatures/Feature>
#include <osgEarthFeatures/FeatureIndex>
#include <osgEarthFeatures/FeatureSourceIndexNode>
#include <osgEarthAnnotation/AnnotationNode>
#include <osgEarthAnnotation/LabelNode>
#include <osgEarthSymbology/Color>
#include <osgEarthSymbology/Style>
#include <osgEarthSymbology/TextSymbol>

#include "Picker.h"
#include "Placard.h"
#include "Renderer.h"
#include "RendererCmd.h"
#include "Trace.h"

using namespace GeoVis;

static const char *highlightVert =
    "#version 130\n"
    "uniform uint objectid_to_highlight; \n"
    "uint oe_index_objectid;      // Stage global containing object id \n"
    "flat out int selected; \n"
    "void checkForHighlight(inout vec4 vertex) \n"
    "{ \n"
    "    selected = (objectid_to_highlight > 1u && objectid_to_highlight == oe_index_objectid) ? 1 : 0; \n"
    "} \n";

static const char *highlightFrag =
    "#version 130\n"
    "flat in int selected; \n"
    "void highlightFragment(inout vec4 color) \n"
    "{ \n"
    "    if ( selected == 1 ) \n"
    "        color.rgb = mix(color.rgb, clamp(vec3(0.5,0.5,2.0)*(1.0-color.rgb), 0.0, 1.0), 0.5); \n"
    "} \n";

static osg::Uniform *s_highlightUniform = NULL;

static void installHighlighter(osg::StateSet *stateSet, int attrLocation)
{
    TRACE("GLSL Version: %s", GLSL_VERSION_STR);
    // This shader program will highlight the selected object.
    osgEarth::VirtualProgram *vp = osgEarth::VirtualProgram::getOrCreate(stateSet);
    vp->setFunction("checkForHighlight",  highlightVert, osgEarth::ShaderComp::LOCATION_VERTEX_CLIP);
    vp->setFunction("highlightFragment",  highlightFrag, osgEarth::ShaderComp::LOCATION_FRAGMENT_COLORING);

    // Since we're accessing object IDs, we need to load the indexing shader as well:
    osgEarth::Registry::objectIndex()->loadShaders(vp);

    // A uniform that will tell the shader which object to highlight:
    if (s_highlightUniform == NULL) {
        s_highlightUniform = new osg::Uniform("objectid_to_highlight", 0u);
    }
    stateSet->addUniform(s_highlightUniform);
}

void GeoVis::setHighlightByObjectID(osgEarth::ObjectID id)
{
    if (s_highlightUniform != NULL) {
        s_highlightUniform->set(id);
    }
}

void GeoVis::clearHighlight()
{
    if (s_highlightUniform != NULL) {
        s_highlightUniform->set(0U);
    }
}

HoverCallback::HoverCallback(Renderer *renderer) :
    _renderer(renderer)
{
    installHighlighter(renderer->getMapNode()->getOrCreateStateSet(),
                       osgEarth::Registry::objectIndex()->getObjectIDAttribLocation());
}

void HoverCallback::onHit(osgEarth::ObjectID id)
{
    // First see whether it's a feature:
    osgEarth::Features::FeatureIndex *index = osgEarth::Registry::objectIndex()->get<osgEarth::Features::FeatureIndex>(id);
    osgEarth::Features::Feature *feature = index ? index->getFeature(id) : 0L;
    TRACE("Hover hit");
    if (feature) {
        TRACE("Hit feature ID: %lu (of %d), oid: %lu", feature->getFID(), index->size(), id);
        const osgEarth::Features::AttributeTable &attrs = feature->getAttrs();
        for (osgEarth::Features::AttributeTable::const_iterator itr = attrs.begin(); itr != attrs.end(); ++itr) {
            TRACE(" attr: %s", itr->first.c_str());
            switch (itr->second.first) {
            case osgEarth::Features::ATTRTYPE_STRING:
                TRACE(" value: %s", itr->second.getString().c_str());
                break;
            case osgEarth::Features::ATTRTYPE_INT:
                TRACE(" value: %d", itr->second.getInt());
                break;
            case osgEarth::Features::ATTRTYPE_DOUBLE:
                TRACE(" value: %g", itr->second.getDouble());
                break;
            case osgEarth::Features::ATTRTYPE_BOOL:
                TRACE(" value: %s", itr->second.getBool() ? "true" : "false");
                break;
            default:
                TRACE(" value: unknown type");
            }
        }
        //TRACE("Feature name: %s", feature->getString("name").c_str());
    } else {
        osgEarth::Annotation::AnnotationNode* anno =
            osgEarth::Registry::objectIndex()->get<osgEarth::Annotation::AnnotationNode>(id);

        std::set<osgEarth::Annotation::AnnotationNode*> toUnHover;
        std::set<osgEarth::Annotation::AnnotationNode*>& hovered = _renderer->getHovered();
        for (std::set<osgEarth::Annotation::AnnotationNode*>::iterator itr = hovered.begin();
             itr != hovered.end(); ++itr) {
            toUnHover.insert(*itr);
        }
        if (anno != NULL) {
            TRACE("Hit AnnotationNode: %p", anno);
#if 0
            if (anno->getDecoration().empty() &&
                hovered.find(anno) == hovered.end()) {
                hovered.insert(anno);
                anno->setDecoration("hover");
                _renderer->eventuallyRender();
            }
            toUnHover.erase(anno);
#endif
        }
#if 0
        for (std::set<osgEarth::Annotation::AnnotationNode *>::iterator itr = toUnHover.begin();
             itr != toUnHover.end(); ++itr) {
            hovered.erase(*itr);
            (*itr)->clearDecoration();
            _renderer->eventuallyRender();
        }
#endif
    }
    setHighlightByObjectID(id);
}

void HoverCallback::onMiss()
{
    TRACE("Hover miss");
    std::set<osgEarth::Annotation::AnnotationNode*> toUnHover;
    std::set<osgEarth::Annotation::AnnotationNode*>& hovered = _renderer->getHovered();
    for (std::set<osgEarth::Annotation::AnnotationNode*>::iterator itr = hovered.begin();
         itr != hovered.end(); ++itr) {
        toUnHover.insert(*itr);
    }
    for (std::set<osgEarth::Annotation::AnnotationNode *>::iterator itr = toUnHover.begin();
         itr != toUnHover.end(); ++itr) {
        hovered.erase(*itr);
#ifndef NEW_ANNOTATION_API
        (*itr)->clearDecoration();
#endif
        _renderer->eventuallyRender();
    }
    clearHighlight();
}

bool HoverCallback::accept(const osgGA::GUIEventAdapter& ea, const osgGA::GUIActionAdapter& aa)
{
    // Pick whenever mouse moves
    return ea.getEventType() == ea.MOVE;
}

SelectCallback::SelectCallback(Renderer *renderer) :
    _renderer(renderer),
    _modifier(false)
{
    installHighlighter(renderer->getMapNode()->getOrCreateStateSet(),
                       osgEarth::Registry::objectIndex()->getObjectIDAttribLocation());
    installHighlighter(renderer->getPlaceNodes()->getOrCreateStateSet(),
                       osgEarth::Registry::objectIndex()->getObjectIDAttribLocation());
}

void SelectCallback::onHit(osgEarth::ObjectID id)
{
    // First see whether it's a feature:
    osgEarth::Features::FeatureIndex *index = osgEarth::Registry::objectIndex()->get<osgEarth::Features::FeatureIndex>(id);
    osgEarth::Features::Feature *feature = index ? index->getFeature(id) : 0L;
    TRACE("Select hit oid: %lu at %g,%g", id, _pickPoint.x(), _pickPoint.y());
    if (feature) {
        osgEarth::Features::FeatureSourceIndex *fsi = dynamic_cast<osgEarth::Features::FeatureSourceIndex *>(index);
        std::string layerName;
        if (fsi) {
            osgEarth::Features::FeatureSource *source = fsi->getFeatureSource();
            if (source) {
                layerName = source->getFeatureSourceOptions().name().get();
            }
        }
        TRACE("Hit FID: %lu (of %d) layer: %s", feature->getFID(), index->size(), layerName.c_str());

        std::vector<unsigned long> fids;
        fids.push_back(feature->getFID());
        _renderer->selectFeatures(fids, layerName.c_str(), !_modifier);
        _renderer->addPlacard(_pickPoint, feature, layerName.c_str());

        char mesg[256];
        if (_modifier) {
            // Check if already selected
            snprintf(mesg, sizeof(mesg), "nv>select feature add {%lu} {%s}\n", feature->getFID(), layerName.c_str());
        } else {
            snprintf(mesg, sizeof(mesg), "nv>select feature set {%lu} {%s}\n", feature->getFID(), layerName.c_str());
        }
        size_t length = strlen(mesg);
        queueResponse(mesg, length, Response::VOLATILE, Response::DATA);
    } else {
        _renderer->clearSelection();
        osgEarth::Annotation::AnnotationNode* anno =
            osgEarth::Registry::objectIndex()->get<osgEarth::Annotation::AnnotationNode>(id);
        if (anno != NULL) {
            TRACE("Hit AnnotationNode: %p, \"%s\"", anno, anno->getName().c_str());
            char mesg[256];
            snprintf(mesg, sizeof(mesg), "nv>select annotation {%s}\n", anno->getName().c_str());
            size_t length = strlen(mesg);
            queueResponse(mesg, length, Response::VOLATILE, Response::DATA);
        }
    }
    setHighlightByObjectID(id);
    _renderer->pickPending(false);
    _renderer->eventuallyRender();
}

void SelectCallback::onMiss()
{
    TRACE("Select miss");
    _renderer->clearSelection();

    static const char *mesg = "nv>select clear\n";
    static const size_t length = strlen(mesg);
    queueResponse(mesg, length, Response::STATIC, Response::DATA);

    _renderer->pickPending(false);
    _renderer->eventuallyRender();
}

bool SelectCallback::accept(const osgGA::GUIEventAdapter& ea, const osgGA::GUIActionAdapter& aa)
{
    if (ea.getEventType() == ea.PUSH) {
        _x = ea.getX();
        _y = ea.getY();
        // Modifier-click?
        _modifier = (ea.getModKeyMask() & osgGA::GUIEventAdapter::MODKEY_CTRL) != 0;
        return false;
    } else if (ea.getEventType() == ea.RELEASE) {
        // Don't pick if mouse was dragged more than 2 pixels
        if (fabsf(ea.getX()-_x) > 2 || fabsf(ea.getY()-_y) > 2) {
            return false;
        } else {
            _renderer->mapMouseCoords(ea.getX(), ea.getY(), _pickPoint, false);
            _renderer->pickPending(true);
            TRACE("Pending pick at %g %g", ea.getX(), ea.getY());
            return true;
        }
    }
    return false;
}

void DeleteCallback::onHit(osgEarth::ObjectID id)
{
    // First see whether it's a feature:
    osgEarth::Features::FeatureIndex *index = osgEarth::Registry::objectIndex()->get<osgEarth::Features::FeatureIndex>(id);
    osgEarth::Features::Feature *feature = index ? index->getFeature(id) : 0L;

    if (feature) {
        TRACE("Hit feature ID: %lu, oid: %lu name: %s", feature->getFID(), id);
        TRACE("Feature name: %s", feature->getString("name").c_str());
    } else {
        TRACE("Picker hit!");
        osgEarth::Annotation::AnnotationNode* anno =
            osgEarth::Registry::objectIndex()->get<osgEarth::Annotation::AnnotationNode>(id);
        std::set<osgEarth::Annotation::AnnotationNode*>& hovered = _renderer->getHovered();
#ifndef NEW_ANNOTATION_API
        anno->clearDecoration();
#endif
        _renderer->getPlaceNodes()->removeChild(anno);
        if (hovered.find(anno) != hovered.end()) {
            hovered.erase(anno);
        }
    }
}
