/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * Copyright (C) 2015  HUBzero Foundation, LLC
 *
 * Author: Leif Delgass <ldelgass@purdue.edu>
 */

#include <string>
#include <sstream>

#include <osg/Depth>
#include <osgText/Text>
#include <osgEarthAnnotation/AnnotationUtils>
#include <osgEarth/Registry>
#include <osgEarth/ShaderGenerator>
#ifdef NEW_ANNOTATION_API
#include <osgEarth/ScreenSpaceLayout>
#include <osgEarthSymbology/BBoxSymbol>
#else
#include <osgEarth/Decluttering>
#endif
#include <osgEarthSymbology/PolygonSymbol>
#include <osgEarthSymbology/TextSymbol>
#include <osgEarthSymbology/Fill>

#include "Placard.h"
#include "Trace.h"

#define LC "[PlacardNode] "

using namespace GeoVis;

Placard::Placard() :
    _enabled(true)
{
    osgEarth::Symbology::TextSymbol *ts = _textStyle.getOrCreateSymbol<osgEarth::Symbology::TextSymbol>();
    //ts->halo()->color() = osgEarth::Symbology::Color(1, 1, 1);
    //ts->halo()->width() = 2;
    ts->fill()->color() = osgEarth::Symbology::Color(0, 0, 0);
    ts->size() = 14;
    ts->alignment() = osgEarth::Symbology::TextSymbol::ALIGN_LEFT_BASE_LINE;
    ts->pixelOffset() = osg::Vec2s(20, 0);
    ts->declutter() = false;
    //ts->encoding() = osgEarth::Symbology::TextSymbol::ENCODING_UTF8;
    double charSize = ts->size().isSet() ? ts->size()->eval() : 14.0;
    _padding = (float)(charSize / 2.0);

#ifdef NEW_ANNOTATION_API
    osgEarth::Symbology::BBoxSymbol *bbs = _textStyle.getOrCreateSymbol<osgEarth::Symbology::BBoxSymbol>();
    bbs->fill() = osgEarth::Symbology::Color(0.75, 0.75, 0.75, 0.75);
    bbs->margin() = _padding;
#else
    // Set backdrop quad color as PolygonSymbol fill
    osgEarth::Symbology::PolygonSymbol *ps = _textStyle.getOrCreateSymbol<osgEarth::Symbology::PolygonSymbol>();
    ps->fill() = osgEarth::Symbology::Color(0.75, 0.75, 0.75, 0.75);
#endif
}

PlacardLabelNode::PlacardLabelNode(osgEarth::MapNode *mapNode,
                                   const osgEarth::GeoPoint& position,
                                   const Placard& placardConf,
                                   const osgEarth::Features::AttributeTable &attrs) :
    osgEarth::Annotation::LabelNode(mapNode, position, placardConf.getStyle()),
    _placardConf(placardConf),
    _attrs(attrs)
{
    setConfig(placardConf);
}

void
PlacardLabelNode::setConfig(const Placard& placardConf)
{
    _placardConf = placardConf;
    std::ostringstream oss;
    for (size_t i = 0; i < placardConf.getNumEntries(); i++) {
        std::string name, label;
        placardConf.getEntry(i, name, label);
        oss << label << ": ";
        TRACE(" attr: %s '%s'", name.c_str(), label.c_str());
        osgEarth::Features::AttributeTable::const_iterator itr = _attrs.find(name);
        if (itr == _attrs.end())
            continue;
        switch (itr->second.first) {
        case osgEarth::Features::ATTRTYPE_STRING:
            TRACE(" value: %s", itr->second.getString().c_str());
            oss << itr->second.getString() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_INT:
            TRACE(" value: %d", itr->second.getInt());
            oss << itr->second.getInt() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_DOUBLE:
            TRACE(" value: %g", itr->second.getDouble());
            oss << itr->second.getDouble() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_BOOL:
            TRACE(" value: %s", itr->second.getBool() ? "true" : "false");
            oss << (itr->second.getBool() ? "true" : "false") << std::endl;
            break;
        default:
            break;
        }
    }
    setText(oss.str());
}

PlacardNode::PlacardNode(osgEarth::MapNode *mapNode,
                         const osgEarth::GeoPoint& position,
                         const Placard& placardConf,
                         const osgEarth::Features::AttributeTable &attrs) :
#ifdef NEW_ANNOTATION_API
    osgEarth::Annotation::GeoPositionNode(mapNode, position),
#else
    osgEarth::Annotation::OrthoNode(mapNode, position),
#endif
    _placardConf(placardConf),
    _attrs(attrs)
{
    setConfig(placardConf);
    init(placardConf.getStyle());
}

void
PlacardNode::setConfig(const Placard& placardConf)
{
    _placardConf = placardConf;
    std::ostringstream oss;
    for (size_t i = 0; i < placardConf.getNumEntries(); i++) {
        std::string name, label;
        placardConf.getEntry(i, name, label);
        oss << label << ": ";
        TRACE(" attr: %s '%s'", name.c_str(), label.c_str());
        osgEarth::Features::AttributeTable::const_iterator itr = _attrs.find(name);
        if (itr == _attrs.end())
            continue;
        switch (itr->second.first) {
        case osgEarth::Features::ATTRTYPE_STRING:
            TRACE(" value: %s", itr->second.getString().c_str());
            oss << itr->second.getString() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_INT:
            TRACE(" value: %d", itr->second.getInt());
            oss << itr->second.getInt() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_DOUBLE:
            TRACE(" value: %g", itr->second.getDouble());
            oss << itr->second.getDouble() << std::endl;
            break;
        case osgEarth::Features::ATTRTYPE_BOOL:
            TRACE(" value: %s", itr->second.getBool() ? "true" : "false");
            oss << (itr->second.getBool() ? "true" : "false") << std::endl;
            break;
        default:
            break;
        }
    }
    setText(oss.str());
}

void
PlacardNode::init( const osgEarth::Symbology::Style& style )
{
#ifdef NEW_ANNOTATION_API
    osgEarth::ScreenSpaceLayout::activate(this->getOrCreateStateSet());
#endif
    _backdropGeode = new osg::Geode();
    _backdropGeode->setComputeBoundingSphereCallback(new osgEarth::Annotation::ControlPointCallback());
#ifdef NEW_ANNOTATION_API
    getPositionAttitudeTransform()->addChild( _backdropGeode.get() );
#else
    getAttachPoint()->addChild(_backdropGeode.get());
#endif
    osg::StateSet* stateSet = _backdropGeode->getOrCreateStateSet();
    stateSet->setAttributeAndModes( new osg::Depth(osg::Depth::ALWAYS, 0, 1, false), 1 );

    _geode = new osg::Geode();

    // ensure that (0,0,0) is the bounding sphere control/center point.
    // useful for things like horizon culling.
    _geode->setComputeBoundingSphereCallback(new osgEarth::Annotation::ControlPointCallback());

#ifdef NEW_ANNOTATION_API
    getPositionAttitudeTransform()->addChild( _geode.get() );
#else
    getAttachPoint()->addChild( _geode.get() );
#endif
    stateSet = _geode->getOrCreateStateSet();
    stateSet->setAttributeAndModes( new osg::Depth(osg::Depth::ALWAYS, 0, 1, false), 1 );

    setStyle( style );
}

void
PlacardNode::setText( const std::string& text )
{
    if ( !_dynamic && getNumParents() > 0 )
    {
        OE_WARN << LC << "Illegal state: cannot change a LabelNode that is not dynamic" << std::endl;
        return;
    }

    _text = text;
    if (!_geode.valid())
        return;

    osgText::Text* d = dynamic_cast<osgText::Text*>(_geode->getDrawable(0));
    if ( d )
    {
        d->setText( text );
        d->dirtyDisplayList();
    }

    if (!_backdropGeode.valid())
        return;

    osg::Geometry *bd = dynamic_cast<osg::Geometry*>(_backdropGeode->getDrawable(0));
    if (bd) {
        osg::BoundingBox bbox = d->getBoundingBox();
        _backdropGeode->removeDrawables( 0, _backdropGeode->getNumDrawables() );
        const osgEarth::Symbology::PolygonSymbol* gs = _style.get<osgEarth::Symbology::PolygonSymbol>();
        osg::Vec4 color(0.75, 0.75, 0.75, 0.75);
        if (gs && gs->fill().isSet()) {
            color = gs->fill().get().color();
        }
        osg::Drawable *bd = osgEarth::Annotation::AnnotationUtils::create2DQuad(bbox, _placardConf.getPadding(), color);
        _backdropGeode->addDrawable(bd);
    }
}

void
PlacardNode::setStyle( const osgEarth::Symbology::Style& style )
{
    if ( !_dynamic && getNumParents() > 0 )
    {
        OE_WARN << LC << "Illegal state: cannot change a LabelNode that is not dynamic" << std::endl;
        return;
    }
#ifndef NEW_ANNOTATION_API
    this->clearDecoration();
#endif
    _backdropGeode->removeDrawables( 0, _backdropGeode->getNumDrawables() );

    _geode->removeDrawables( 0, _geode->getNumDrawables() );

    _style = style;

    const osgEarth::Symbology::TextSymbol* symbol = _style.get<osgEarth::Symbology::TextSymbol>();

    if ( _text.empty() )
        _text = symbol->content()->eval();

    osg::Drawable* t = osgEarth::Annotation::AnnotationUtils::createTextDrawable( _text, symbol, osg::Vec3(0,0,0) );
    _geode->addDrawable(t);
    _geode->setCullingActive(false);

    const osgEarth::Symbology::PolygonSymbol* gs = _style.get<osgEarth::Symbology::PolygonSymbol>();
    osg::Vec4 color(0.75, 0.75, 0.75, 0.75);
    if (gs && gs->fill().isSet()) {
        color = gs->fill().get().color();
    }

    osg::BoundingBox bbox = t->getBoundingBox();
    osg::Drawable *bd = osgEarth::Annotation::AnnotationUtils::create2DQuad(bbox, _placardConf.getPadding(), color);
    _backdropGeode->addDrawable(bd);
    _backdropGeode->setCullingActive(false);

    applyStyle( _style );

    setLightingIfNotSet( false );

    osgEarth::Registry::shaderGenerator().run(
        this,
        "osgEarth.PlacardNode",
        osgEarth::Registry::stateSetCache() );
#ifdef NEW_ANNOTATION_API
    // Set to max priority to draw on top in decluttering sort
    setPriority(FLT_MAX);
    // setPriority calls updateLayoutData
    //updateLayoutData();
#endif
}

#ifdef NEW_ANNOTATION_API
void
PlacardNode::setPriority(float value)
{
    osgEarth::Annotation::GeoPositionNode::setPriority(value);
    updateLayoutData();
}

void
PlacardNode::updateLayoutData()
{
    if (!_geode.valid())
        return;

    if (!_dataLayout.valid())
    {
        _dataLayout = new osgEarth::ScreenSpaceLayoutData();
    }

    // re-apply annotation drawable-level stuff as neccesary.
    for (unsigned i = 0; i < _backdropGeode->getNumDrawables(); ++i)
    {
        _backdropGeode->getDrawable(i)->setUserData(_dataLayout.get());
    }
    // re-apply annotation drawable-level stuff as neccesary.
    for (unsigned i = 0; i < _geode->getNumDrawables(); ++i)
    {
        _geode->getDrawable(i)->setUserData(_dataLayout.get());
    }

    _dataLayout->setPriority(getPriority());
    const osgEarth::Symbology::TextSymbol *ts = getStyle().get<osgEarth::Symbology::TextSymbol>();
    if (ts)
    {
        _dataLayout->setPixelOffset(ts->pixelOffset().get());
    }
}
#else
void
PlacardNode::setAnnotationData( osgEarth::Annotation::AnnotationData* data )
{
    osgEarth::Annotation::OrthoNode::setAnnotationData( data );

    if (!_geode.valid())
        return;

    // override this method so we can attach the anno data to the drawables.
    for(unsigned i=0; i<_geode->getNumDrawables(); ++i)
    {
        _geode->getDrawable(i)->setUserData( data );
    }
}
#endif

void
PlacardNode::setDynamic( bool dynamic )
{
#ifdef NEW_ANNOTATION_API
    osgEarth::Annotation::GeoPositionNode::setDynamic( dynamic );
#else
    osgEarth::Annotation::OrthoNode::setDynamic( dynamic );
#endif
    osgText::Text* d = dynamic_cast<osgText::Text*>(_geode->getDrawable(0));
    if ( d )
    {
        d->setDataVariance( dynamic ? osg::Object::DYNAMIC : osg::Object::STATIC );
    }
    osg::Geometry *bd = dynamic_cast<osg::Geometry*>(_backdropGeode->getDrawable(0));
    if (bd) {
        bd->setDataVariance( dynamic ? osg::Object::DYNAMIC : osg::Object::STATIC );
    }
}

osgEarth::Config
PlacardNode::getConfig() const
{
#ifdef NEW_ANNOTATION_API
    osgEarth::Config conf = osgEarth::Annotation::GeoPositionNode::getConfig();
#else
    osgEarth::Config conf = osgEarth::Annotation::OrthoNode::getConfig();
#endif
    conf.key() = "placard";

    conf.add   ( "text",   _text );
    conf.addObj( "style",  _style );
    conf.add   ( "padding", _placardConf.getPadding() );

    return conf;
}
