/* -*- mode: c++; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * Copyright (C) 2015  HUBzero Foundation, LLC
 *
 * Author: Leif Delgass <ldelgass@purdue.edu>
 */

#ifndef GEOVIS_PLACARD_H
#define GEOVIS_PLACARD_H

#include <string>
#include <vector>
#include <algorithm>

#ifdef NEW_ANNOTATION_API
#include <osgEarth/ScreenSpaceLayout>
#include <osgEarthAnnotation/GeoPositionNode>
#else
#include <osgEarthAnnotation/Decoration>
#include <osgEarthAnnotation/OrthoNode>
#endif
#include <osgEarthAnnotation/LabelNode>
#include <osgEarthFeatures/Feature>

namespace GeoVis {

class Placard {
public:
    Placard();
    virtual ~Placard()
    {}
    void addAllAttributes(const osgEarth::Features::AttributeTable& attrs)
    {
        for (osgEarth::Features::AttributeTable::const_iterator itr = attrs.begin();
             itr != attrs.end(); ++itr) {
            std::string label = itr->first;
            std::replace(label.begin(), label.end(), '_', ' ');
            addEntry(itr->first, label);
        }
    }
    void addEntry(const std::string& name)
    {
        addEntry(name, name);
    }
    void addEntry(const std::string& name, const std::string& label)
    {
        _attrNames.push_back(name);
        _attrLabels.push_back(label);
    }
    void getEntry(int i, std::string& name, std::string& label) const
    {
        name = _attrNames[i];
        label = _attrLabels[i];
    }
    size_t getNumEntries() const
    {
        return _attrNames.size();
    }
    void setStyle(const osgEarth::Symbology::Style& style)
    {
        _textStyle = style;
    }
    const osgEarth::Symbology::Style& getStyle() const
    {
        return _textStyle;
    }
    void setPadding(float padding)
    {
        _padding = padding;
    }
    float getPadding() const
    {
        return _padding;
    }
    bool enabled()
    {
        return _enabled;
    }
    void setEnabled(bool state)
    {
        _enabled = state;
    }

private:
    std::vector<std::string> _attrNames;
    std::vector<std::string> _attrLabels;
    float _padding;
    osgEarth::Symbology::Style _textStyle;
    bool _enabled;
};

class PlacardLabelNode : public osgEarth::Annotation::LabelNode {
public:
    META_Node(osgEarthAnnotation, PlacardLabelNode);
#ifndef NEW_ANNOTATION_API
    virtual bool accept(osgEarth::Annotation::Decoration* ds, bool enable)
    { return ds->apply(*this, enable); }
#endif
    PlacardLabelNode() :
        osgEarth::Annotation::LabelNode()
    {}
    PlacardLabelNode(const PlacardLabelNode& other, const osg::CopyOp& op = osg::CopyOp::DEEP_COPY_ALL) :
        osgEarth::Annotation::LabelNode(other, op),
        _placardConf(other._placardConf)
    {
    }
    PlacardLabelNode(osgEarth::MapNode *mapNode,
                     const osgEarth::GeoPoint& position,
                     const Placard& placardConf,
                     const osgEarth::Features::AttributeTable &attrs);

    virtual ~PlacardLabelNode()
    {}

    void setConfig(const Placard& placardConf);

private:
    Placard _placardConf;
    osgEarth::Features::AttributeTable _attrs;
};

class PlacardNode :
#ifdef NEW_ANNOTATION_API
public osgEarth::Annotation::GeoPositionNode
#else
public osgEarth::Annotation::OrthoNode
#endif
{
public:
    META_Node(osgEarthAnnotation, PlacardNode);
#ifndef NEW_ANNOTATION_API
    virtual bool accept(osgEarth::Annotation::Decoration* ds, bool enable)
    { return ds->apply(*this, enable); }
#endif
    PlacardNode() :
#ifdef NEW_ANNOTATION_API
        osgEarth::Annotation::GeoPositionNode()
#else
        osgEarth::Annotation::OrthoNode()
#endif
    {}
    PlacardNode(const PlacardNode& other, const osg::CopyOp& op = osg::CopyOp::DEEP_COPY_ALL) :
#ifdef NEW_ANNOTATION_API
        osgEarth::Annotation::GeoPositionNode(other, op),
#else
        osgEarth::Annotation::OrthoNode(other, op),
#endif
        _placardConf(other._placardConf)
    {
    }
    PlacardNode(osgEarth::MapNode *mapNode,
                const osgEarth::GeoPoint& position,
                const Placard& placardConf,
                const osgEarth::Features::AttributeTable &attrs);

    virtual ~PlacardNode()
    {}

    void setConfig(const Placard& placardConf);

    /**
     * Sets the text content.
     */
    void setText( const std::string& text );
    const std::string& text() const { return _text; }
    virtual const std::string& getText() const { return text(); }

    /**
     * Gets a copy of the text style.
     */
    const osgEarth::Symbology::Style& getStyle() const { return _style; }

    /**
     * Sets a new text style
     */
    void setStyle( const osgEarth::Symbology::Style& style );

public: // Ortho|GeoPositionNode override
#ifdef NEW_ANNOTATION_API
    virtual void setPriority(float value);
#else
    virtual void setAnnotationData( osgEarth::Annotation::AnnotationData* data );
#endif
    virtual void setDynamic( bool value );

    virtual osgEarth::Config getConfig() const;

protected:
    void init(const osgEarth::Symbology::Style& style);
#ifdef NEW_ANNOTATION_API
    void updateLayoutData();
#endif

    std::string                _text;
    osgEarth::Symbology::Style _style;
    osg::ref_ptr<osg::Geode>   _geode;
    osg::ref_ptr<osg::Geode>   _backdropGeode;
#ifdef NEW_ANNOTATION_API
    osg::ref_ptr<osgEarth::ScreenSpaceLayoutData> _dataLayout;
#endif

    Placard _placardConf;
    osgEarth::Features::AttributeTable _attrs;
};

}

#endif
